/*
    BNoise - A 2D noise generator.
    Copyright (C) 2023 Pascal Baillehache info@baillehachepascal.dev
    https://baillehachepascal.dev
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
    You should have received a copy of the GNU General Public License
    along with this program. If not, see <http://www.gnu.org/licenses/>.
*/
#include "bnoise.h"

// Shortcut equivalent to a for loop iterating from 0 to (nbIter-1)
// and storing the current iteration index in varName
#define loop(varName, nbIter)                 \
  for(__typeof__((nbIter) + 0) (varName) = 0; \
       (varName) < (nbIter); ++(varName))

// Get a pseudo random number in [0,1]
// Output:
//   Return the pseudo random number.
static float rng(void) {
  return (float)rand() / (float)RAND_MAX;
}

// Allocate memory and create a BNoise instance
// Inputs:
//   order: order of the noise. The grid will be of size (dim+1)*(dim+1) in
//        number of vertices and the Bezier curve will be of order (4+dim).
// Output:
//   Return the new BNoise instance
BNoise* BNoiseAlloc(int const order) {

  // Allocate memory
  BNoise* that = malloc(sizeof(BNoise));

  // Set the dimension of the grid
  that->dim = order + 1;
  that->fDim = (float)(that->dim);

  // Allocate memory for the vertices value and initialise them to random
  // number in [0,1]
  size_t nbVertex = that->dim * that->dim;
  that->map = malloc(nbVertex * sizeof(float));
  loop(i, nbVertex) that->map[i] = rng();

  // Calculate the number of controls in a spline
  int nbCtrl = that->dim * 3 + 1;

  // Loop on the four bezier splines
  loop(iCurve, 2) loop(iCoord, 2) {

    // Allocate memory for the bezier spline controls (add 3 for an extra
    // segment, cf below)
    that->curve[iCurve][iCoord] = malloc(sizeof(float) * (nbCtrl + 3));

    // The handles are set to random numbers in [0, order]
    float* ctrls = that->curve[iCurve][iCoord];
    loop(iCtrl, nbCtrl) {
      if((iCtrl % 3) != 0) ctrls[iCtrl] = rng() * (float)order;
    }

    // Calculate anchors as average of their neighbour handles and avoid the
    // degenerate case ctrl_i == ctrl_{i-2}
    loop(iCtrl, nbCtrl) {
      if((iCtrl % 3) == 0) {
        int iCtrlPrev = (iCtrl - 1) % (nbCtrl - 1);
        if(iCtrlPrev < 0) iCtrlPrev += nbCtrl - 1;
        int iCtrlNext = (iCtrl + 1) % (nbCtrl - 1);
        if(fabs(ctrls[iCtrlPrev] - ctrls[iCtrlNext]) < 0.1) {
          ctrls[iCtrlPrev] = fmodf(ctrls[iCtrlNext] + 0.1, order);
        }
        ctrls[iCtrl] = 0.5 * (ctrls[iCtrlPrev] + ctrls[iCtrlNext]);
      }
    }

    // Duplicate the first segment at the end to avoid a modulo in BNoiseGet()
    loop(iCtrl, 3) ctrls[nbCtrl + iCtrl] = ctrls[1 + iCtrl];
  }

  // Return the new noise instance
  return that;
}

// Free memory used by a BNoise
// Inputs:
//   that: the BNoise
void BNoiseFree(BNoise** const that) {
  if(that == NULL || *that == NULL) return;
  loop(iCurve, 2) loop(iCoord, 2) free((*that)->curve[iCurve][iCoord]);
  free((*that)->map);
  free(*that);
  *that = NULL;
}

// Smoother step function
// Inputs:
//   x: the input value in [0,1]
// Output:
//   Return the smoothed value, in [0, 1]
static float SmootherStep(float const x) {
  return x * x * x * (x * (x * 6.0 - 15.0) + 10.0);
}

// LERP function, map a float value linearly from [0,1] to a range
// Inputs:
//      x: the input
//     to: range of the output
// Output:
//   Return the mapped input
static float Lerp(
         float const x,
  float const* const to) {
  return to[0] + x * (to[1] - to[0]);
}

// Calculate the value of a Bezier curve
// Inputs:
//   t: argument of the function, in [0, 1]
//   params: the control values of the Bezier
//   order: the order of the Bezier
// Output:
//   Return the value of the Bezier
static float EvalBezier(
         float const t,
  float const* const params,
           int const order) {
  float v[order + 1];
  memcpy(v, params, sizeof(float) * (order + 1));
  loop(depth, order) loop(i, order - depth) v[i] = Lerp(t, v + i);
  return v[0];
}

// Get the noise value at a given position
// Inputs:
//   that: the BNoise
//   pos: the 2D position (one tile is contained in [0,1]x[0,1])
// Output:
//   Return the noise value at the position, in [0,1]
float BNoiseGet(
  BNoise* const that,
    float const pos[2]) {

  // Warp and scale the input position from [0, 1] over the interval
  // [0, that->dim]
  float modPos[2] = {0.0};
  float t[2] = {0.0};
  int idxCurve[2] = {0};

  loop(iCoord, 2) {
    if(pos[iCoord] >= 0.0) {
      modPos[iCoord] = that->fDim * fmodf(pos[iCoord], 1.0);
    } else {
      modPos[iCoord] = that->fDim * (1.0 - fmodf(fabs(pos[iCoord]), 1.0));
    }
    t[iCoord] = fmodf(modPos[iCoord], 1.0);
    idxCurve[iCoord] = floor(modPos[iCoord]);
  }

  // Map the warped position to a grid position using the two curve beziers
  float mappedPos[2] = {0.0};
  loop(iCoord, 2) {
    float curvePos[2] = {0.0};
    loop(iCurve, 2) {

      // Switch the coordinates
      int idxCoord = (iCoord == iCurve);

      // Calculate the position in the grid for the curve bezier given the
      // input calculated above
      curvePos[iCurve] =
        EvalBezier(
          t[idxCoord], that->curve[iCurve][iCoord] + idxCurve[idxCoord] * 3, 3);
    }

    // Get the mapped position as the center of the positions given by the
    // beziers curves
    mappedPos[iCoord] = 0.5 * (curvePos[0] + curvePos[1]);
  }

  // Get the indices of the cell and local position in that cell for the
  // mapped position
  int posIdx[2] = {0};
  float posCell[2] = {0.0};
  loop(iCoord, 2) {
    posIdx[iCoord] = (int)floor(mappedPos[iCoord]);
    posCell[iCoord] = fmodf(mappedPos[iCoord], 1.0);

    // Apply smootherstep to the local position for derivability
    posCell[iCoord] = SmootherStep(posCell[iCoord]);
  }

  // Calculate the result value using bilinear interpolation on the cell
  // corners value using the local position
  float val[2] = {0.0};
  loop(i, 2) {
    val[i] =
      Lerp(posCell[0], that->map + (posIdx[1] + i) * that->dim + posIdx[0]);
  }
  return Lerp(posCell[1], val);
}
